/* di.cpp - diskinfo for BeOS. A crossover between BeOS df and
 * AmigaOS Info commands.
 *
 * di (c) 2004 by Christian Packmann <Christan.Packmann@gmx.de>
 * 
 * This program is free software you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#include <iostream>
#include <string>
#include <list>
#include <algorithm>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <StorageKit.h>
#include <fs_info.h>
#include <Node.h>
#include <Drivers.h>
#include <unistd.h>

// Volume info storage, used in a list<VolInfo> to allow sorting
// prior to printing the information.
class VolInfo {
private:
	string mName;
	string mFSType;
	int32 mBlockSize;
	double mCapacity;
	double mFree;
	double mFreePercent;
	double mUsed;
	double mUsedPercent;
	bool mPersistent;
	bool mReadOnly;
	bool mRemovable;
	bool mShared;
	bool mAttr;
	bool mMime;
	bool mQuery;
	string mFlagsString;
	char mDateStr[25];
public:
	void SetTo(BVolume* vol);
	// read attributes
	const char *Name(void) { return mName.c_str(); };
	double Capacity(void) { return mCapacity; };
	double Free(void) { return mFree; };
	double FreePercent(void) { return mFreePercent; };
	double Used(void) { return mUsed; };
	double UsedPercent(void) { return mUsedPercent; };
	const char *Flags(void) { return mFlagsString.c_str(); };
	const char *FSType(void) { return mFSType.c_str(); };
	int32 BlockSize(void) { return mBlockSize; };
	char *Date(void) { return mDateStr; };
	// "<"-operator for using sort() in containers.
	bool operator<(const VolInfo vi) {
		return (strcasecmp(mName.c_str(), vi.mName.c_str()) < 0) ? true : false;
	};
};
void VolInfo::SetTo(BVolume* vol) {
	// Get Volume name
	char name[B_FILE_NAME_LENGTH];
	vol->GetName(name);
	mName = name;
	// Get filesystem type & flags
	dev_t dev = vol->Device();
	fs_info fsinfo;
	if(0 == fs_stat_dev(dev, &fsinfo) ) {
		mFSType = fsinfo.fsh_name;
		mBlockSize = fsinfo.block_size;
		mQuery = (fsinfo.flags & B_FS_HAS_QUERY);
		mAttr = (fsinfo.flags & B_FS_HAS_ATTR);
		mMime = (fsinfo.flags & B_FS_HAS_MIME);
		mPersistent = (fsinfo.flags & B_FS_IS_PERSISTENT);
		mRemovable = (fsinfo.flags & B_FS_IS_REMOVABLE);
		mReadOnly = (fsinfo.flags & B_FS_IS_READONLY);
		mShared = (fsinfo.flags & B_FS_IS_SHARED);
		// Read Removable flag low-level, to work around bug in various
		// filesystems
		int fd = open(fsinfo.device_name, O_RDONLY | O_CLOEXEC);
		if(fd >= 0) {
			device_geometry geom;
			if(B_OK <= ioctl(fd, B_GET_GEOMETRY, &geom, sizeof(device_geometry))) {
				if(geom.removable)
					mRemovable = true;
			}
			else
				printf("ioctl() failed for %s\n", name);
		}
		else
			printf("Can't open device for %s\n", name);
	}
	BDirectory volDir;
	BPath volPath;
	if((B_OK == vol->GetRootDirectory(&volDir)) &&
		B_OK == volPath.SetTo(&volDir, NULL))
	{
		// Generate date string
		time_t t;
		volDir.GetCreationTime(&t);
		struct tm* date = localtime(&t);
		if(date)
			strftime(mDateStr, sizeof(mDateStr), "%F", date);
		else
			mDateStr[0] = 0;
	}
	else
		printf("Can't get dir/path for %s\n", name);
	// Generate flags string
	// - Order of flags is taken from df command.
	// - Flags whose position is unknown (Shared, Removable) are currently
	//   included as an empty space in the returned string.
	mQuery = vol->KnowsQuery();
	mAttr = vol->KnowsAttr();
	mMime = vol->KnowsMime();
	mPersistent = vol->IsPersistent();
	mReadOnly = vol->IsReadOnly();
	mFlagsString = "";
	mFlagsString += mQuery ? "Q" : "-";
	mFlagsString += mAttr ? "A" : "-";
	mFlagsString += mMime ? "M" : "-";
	mFlagsString += mShared ? "S" : "-";
	mFlagsString += mPersistent ? "P" : "-";
	mFlagsString += mRemovable ? "R" : "-";
	mFlagsString += mReadOnly ? "-" : "W";
	// Get capacity, calcuate used space & percentages
	mCapacity = (double)vol->Capacity() / (1024.0*1024*1024);
	mFree = (double)vol->FreeBytes() / (1024.0*1024*1024);
	if(mCapacity > 0) {
		mFreePercent = (mFree/mCapacity)*100;
		mUsed = mCapacity - mFree;
		mUsedPercent = (mUsed/mCapacity)*100;
	}
}

int main(void) {
	BVolumeRoster volRoster;
	BVolume* vol;
	BDirectory* dir;
	VolInfo vi;
	list<VolInfo> viList;
	//struct tm date;
	vol = new BVolume();
	dir = new BDirectory();
	while( B_OK == volRoster.GetNextVolume(vol) ) {
		char name[B_FILE_NAME_LENGTH];
		double capacity, freeBytes, freePercent;
		vol->GetName(name);
		if(0!=strcmp(name,"")) { // Eh, bug: should simply check IsPersistent() for our purposes.
			vi.SetTo(vol);
			viList.push_back(vi);
		}
	}
	list<VolInfo>::iterator viIter;
	viList.sort();
	printf("%-20s %-6s    %-14s %-7s %-8s %-2s %-10s\n",
		"        Volume", "  Size", "     Free", " Flags",
		"  Type", "BS", " Created");
	for(viIter=viList.begin(); viIter!=viList.end(); ++viIter) {
		// Get name string, truncate to length=20 if necessary
		string name = viIter->Name();
		if(name.length() > 20)
			name.resize(20);
		printf("%-20s %6.2fGB, %6.2fGB (%2.0f%%) %7s %8s %1dK %10s\n",
			name.c_str(),
			viIter->Capacity(), viIter->Free(),	viIter->FreePercent(),
			viIter->Flags(), viIter->FSType(), viIter->BlockSize()/1024, 
			viIter->Date() );
	}
}
